/* $Id: upnpsoap.c,v 1.87 2011/07/15 07:48:26 nanard Exp $ */
/* MiniUPnP project
 * http://miniupnp.free.fr/ or http://miniupnp.tuxfamily.org/
 * (c) 2006-2010 Thomas Bernard 
 * This software is subject to the conditions detailed
 * in the LICENCE file provided within the distribution */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/socket.h>
#include <unistd.h>
#include <syslog.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <netdb.h>

#include "config.h"
#include "upnpglobalvars.h"
#include "upnphttp.h"
#include "upnpsoap.h"
#include "upnpreplyparse.h"
#include "upnpredirect.h"
#include "getifstats.h"
#include "getconnstatus.h"
#include "upnpurns.h"

static void
BuildSendAndCloseSoapResp(struct upnphttp * h,
                          const char * body, int bodylen)
{
	static const char beforebody[] =
		"<?xml version=\"1.0\"?>\r\n"
		"<s:Envelope xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" "
		"s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">"
		"<s:Body>";

	static const char afterbody[] =
		"</s:Body>"
		"</s:Envelope>\r\n";

	BuildHeader_upnphttp(h, 200, "OK",  sizeof(beforebody) - 1
		+ sizeof(afterbody) - 1 + bodylen );

	memcpy(h->res_buf + h->res_buflen, beforebody, sizeof(beforebody) - 1);
	h->res_buflen += sizeof(beforebody) - 1;

	memcpy(h->res_buf + h->res_buflen, body, bodylen);
	h->res_buflen += bodylen;

	memcpy(h->res_buf + h->res_buflen, afterbody, sizeof(afterbody) - 1);
	h->res_buflen += sizeof(afterbody) - 1;

	SendResp_upnphttp(h);
	CloseSocket_upnphttp(h);
}

static void
GetConnectionTypeInfo(struct upnphttp * h, const char * action, int v2)
{
	static const char resp[] =
		"<u:GetConnectionTypeInfoResponse "
		"xmlns:u=\"" SERVICE_TYPE_WANIPC "%d\">"
		"<NewConnectionType>IP_Routed</NewConnectionType>"
		"<NewPossibleConnectionTypes>IP_Routed</NewPossibleConnectionTypes>"
		"</u:GetConnectionTypeInfoResponse>";
	char body[512];
	int bodylen;

	action = action;
	bodylen = snprintf(body, sizeof(body), resp, v2 ? 2 : 1);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetTotalBytesSent(struct upnphttp * h, const char * action, int v2)
{
	int r;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewTotalBytesSent>%lu</NewTotalBytesSent>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct ifdata data;

	v2 = v2;
	r = getifstats(ext_if_name, &data);
	bodylen = snprintf(body, sizeof(body), resp,
	         action, "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
             r<0?0:data.obytes, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetTotalBytesReceived(struct upnphttp * h, const char * action, int v2)
{
	int r;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewTotalBytesReceived>%lu</NewTotalBytesReceived>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct ifdata data;

	v2 = v2;
	r = getifstats(ext_if_name, &data);
	bodylen = snprintf(body, sizeof(body), resp,
	         action, "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
	         r<0?0:data.ibytes, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetTotalPacketsSent(struct upnphttp * h, const char * action, int v2)
{
	int r;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewTotalPacketsSent>%lu</NewTotalPacketsSent>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct ifdata data;

	v2 = v2;
	r = getifstats(ext_if_name, &data);
	bodylen = snprintf(body, sizeof(body), resp,
	         action, "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
	         r<0?0:data.opackets, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetTotalPacketsReceived(struct upnphttp * h, const char * action, int v2)
{
	int r;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<NewTotalPacketsReceived>%lu</NewTotalPacketsReceived>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct ifdata data;

	v2 = v2;
	r = getifstats(ext_if_name, &data);
	bodylen = snprintf(body, sizeof(body), resp,
	         action, "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
	         r<0?0:data.ipackets, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetCommonLinkProperties(struct upnphttp * h, const char * action, int v2)
{
	/* WANAccessType : set depending on the hardware :
	 * DSL, POTS (plain old Telephone service), Cable, Ethernet */
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		/*"<NewWANAccessType>DSL</NewWANAccessType>"*/
		"<NewWANAccessType>Cable</NewWANAccessType>"
		"<NewLayer1UpstreamMaxBitRate>%lu</NewLayer1UpstreamMaxBitRate>"
		"<NewLayer1DownstreamMaxBitRate>%lu</NewLayer1DownstreamMaxBitRate>"
		"<NewPhysicalLinkStatus>%s</NewPhysicalLinkStatus>"
		"</u:%sResponse>";

	char body[2048];
	int bodylen;
	struct ifdata data;
	const char * status = "Up";	/* Up, Down (Required),
	                             * Initializing, Unavailable (Optional) */

	v2 = v2;
	if((downstream_bitrate == 0) || (upstream_bitrate == 0))
	{
		if(getifstats(ext_if_name, &data) >= 0)
		{
			if(downstream_bitrate == 0) downstream_bitrate = data.baudrate;
			if(upstream_bitrate == 0) upstream_bitrate = data.baudrate;
		}
	}
	if (external_addr == 0)
		status = "Down";
	bodylen = snprintf(body, sizeof(body), resp,
	    action, "urn:schemas-upnp-org:service:WANCommonInterfaceConfig:1",
		upstream_bitrate, downstream_bitrate,
	    status, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetStatusInfo(struct upnphttp * h, const char * action, int v2)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s%d\">"
		"<NewConnectionStatus>%s</NewConnectionStatus>"
		"<NewLastConnectionError>ERROR_NONE</NewLastConnectionError>"
		"<NewUptime>%ld</NewUptime>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	time_t uptime;
	const char * status;
	/* ConnectionStatus possible values :
	 * Unconfigured, Connecting, Connected, PendingDisconnect,
	 * Disconnecting, Disconnected */

	status = get_wan_connection_status_str(ext_if_name);
	uptime = (time(NULL) - startup_time);
	bodylen = snprintf(body, sizeof(body), resp,
		action, SERVICE_TYPE_WANIPC, v2 ? 2 : 1,
		status, (long)uptime, action);	
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetNATRSIPStatus(struct upnphttp * h, const char * action, int v2)
{
	static const char resp[] =
		"<u:GetNATRSIPStatusResponse "
		"xmlns:u=\"" SERVICE_TYPE_WANIPC "%d\">"
		"<NewRSIPAvailable>0</NewRSIPAvailable>"
		"<NewNATEnabled>1</NewNATEnabled>"
		"</u:GetNATRSIPStatusResponse>";
	char body[512];
	int bodylen;
	/* 2.2.9. RSIPAvailable
	 * This variable indicates if Realm-specific IP (RSIP) is available
	 * as a feature on the InternetGatewayDevice. RSIP is being defined
	 * in the NAT working group in the IETF to allow host-NATing using
	 * a standard set of message exchanges. It also allows end-to-end
	 * applications that otherwise break if NAT is introduced
	 * (e.g. IPsec-based VPNs).
	 * A gateway that does not support RSIP should set this variable to 0. */
	action = action;
	bodylen = snprintf(body, sizeof(body), resp, v2 ? 2 : 1);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
GetExternalIPAddress(struct upnphttp *h, const char *action, int v2)
{
	static const char resp[] =
		"<u:GetExternalIPAddressResponse "
		"xmlns:u=\"urn:schemas-upnp-org:service:WANIPConnection:%d\">"
		"<NewExternalIPAddress>%s</NewExternalIPAddress>"
		"</u:GetExternalIPAddressResponse>";
	char body[512];
	int bodylen;
	struct in_addr in;
	/* Does that method need to work with IPv6 ?
	 * There is usually no NAT with IPv6 */

	action = action;
	if (external_addr != 0)
		in.s_addr = external_addr;
	else
		in.s_addr = 0;
	bodylen = snprintf(body, sizeof(body), resp,
			   v2 ? 2 : 1, inet_ntoa(in));
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

void
AddPortMappingHandler(struct upnphttp *h, int v2)
{
	static const char resp[] =
		"<u:AddPortMappingResponse "
		"xmlns:u=\"" SERVICE_TYPE_WANIPC "%d\"/>";
	char body[512];
	int bodylen;

	bodylen = snprintf(body, sizeof(body), resp, v2 ? 2 : 1);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

void
AddAnyPortMappingHandler(struct upnphttp *h, unsigned short eport)
{
	static const char resp[] =
		"<u:AddAnyPortMappingResponse "
		"xmlns:u=\"" SERVICE_TYPE_WANIPC "2\">"
		"<NewReservedPort>%hu</NewReservedPort>"
		"</u:AddAnyPortMappingResponse>";
	char body[512];
	int bodylen;

	bodylen = snprintf(body, sizeof(body), resp, eport);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

/* possible error codes for AddPortMapping :
 * 402 - Invalid Args
 * 501 - Action Failed
 * 606 - Action not authorized (v2)
 * 715 - Wildcard not permited in InternalClient
 * 716 - Wildcard not permited in ExternalPort
 * 718 - ConflictInMappingEntry
 * 724 - SamePortValuesRequired (deprecated in IGD v2)
 * 725 - OnlyPermanentLeasesSupported
 *	 The NAT implementation only supports permanent lease
 *	 times on port mappings (deprecated in IGD v2)
 * 726 - RemoteHostOnlySupportsWildcard
 *	 RemoteHost must be a wildcard and cannot be a
 *	 specific IP address or DNS name (deprecated in IGD v2)
 * 727 - ExternalPortOnlySupportsWildcard
 *	 ExternalPort must be a wildcard and cannot be a
 *	 specific port value (deprecated in IGD v2)
 * 728 - NoPortMapsAvailable
 *       There are not enough free prots available to complete
 *       the mapping (added in IGD v2)
 * 729 - ConflictWithOtherMechanisms
 *       Attempted port mapping is not allowed due to
 *       conflict with other mechanisms (added in IGD v2)
 * 732 - WildCardNotPermittedInIntPort
 *       InternalPort cannot be wild-carded (added in IGD v2)
 * Ignored argument : NewEnabled */
static void
AddPortMapping(struct upnphttp *h, const char *action, int v2)
{
	struct NameValueParserData data;
	char *int_ip, *int_port, *ext_port, *protocol, *desc;
	char *r_host, *leaseduration_str;
	unsigned short iport, eport;
	unsigned int leaseduration;
	struct hostent *hp;
	char **ptr;
	struct in_addr result_ip, remote;

	if (external_addr == 0) {
		SoapError(h, 501, "ActionFailed");
		return;
	}

	ParseNameValue(h->req_buf + h->req_contentoff,
		       h->req_contentlen, &data);

	/* IGD 2 MUST support both wildcard and specific IP address
	 * values for RemoteHost (only the wildcard value was REQUIRED
	 * in release 1.0) */
	r_host = GetValueFromNameValueList(&data, "NewRemoteHost");

	int_ip = GetValueFromNameValueList(&data, "NewInternalClient");
	if (int_ip == NULL) {
		ClearNameValueList(&data);
		SoapError(h, 715, "WildCardNotPermittedInSrcIP");
		return;
	}

	/* if ip not valid assume hostname and convert */
	if (inet_pton(AF_INET, int_ip, &result_ip) <= 0) {
		hp = gethostbyname(int_ip);
		if (hp && (hp->h_addrtype == AF_INET))
			for (ptr = hp->h_addr_list;
			     (ptr != NULL) && (*ptr != NULL);
			     ptr++) {
				int_ip = inet_ntoa(*((struct in_addr *) *ptr));
				result_ip = *((struct in_addr *) *ptr);
				/* TODO : deal with more than
				   one ip per hostname */
				break;
			}
		else {
			syslog(LOG_ERR,
			       "Failed to convert hostname '%s' to ip address",
			       int_ip); 
			ClearNameValueList(&data);
			SoapError(h, 402, "Invalid Args");
			return;
		}				
	}

	/* if rhost not valid assume hostname and convert */
	if (r_host != NULL) {
		if (inet_pton(AF_INET, r_host, &remote) <= 0) {
			hp = gethostbyname(r_host);
			if (hp && (hp->h_addrtype == AF_INET))
				for (ptr = hp->h_addr_list;
				     (ptr != NULL) && (*ptr != NULL);
				     ptr++) {
					remote = *((struct in_addr *) *ptr);
					r_host = inet_ntoa(remote);
					/* TODO : deal with more than
					   one ip per hostname */
					break;
				}
			else {
				syslog(LOG_ERR,
				       "Failed to convert hostname '%s'"
				       "to ip address",
				       r_host); 
				ClearNameValueList(&data);
				SoapError(h, 402, "Invalid Args");
				return;
			}
		}
	}

	/* check if NewInternalAddress is the client address */
	if (GETFLAG(SECUREMODEMASK))
		if (h->clientaddr.s_addr != result_ip.s_addr) {
			syslog(LOG_INFO,
			       "Client %s tried to redirect port to %s",
			       inet_ntoa(h->clientaddr), int_ip);
			ClearNameValueList(&data);
			if (v2)
				SoapError(h, 606, "Action not authorized");
			else
				SoapError(h, 718, "ConflictInMappingEntry");
			return;
		}

	int_port = GetValueFromNameValueList(&data, "NewInternalPort");
	ext_port = GetValueFromNameValueList(&data, "NewExternalPort");
	protocol = GetValueFromNameValueList(&data, "NewProtocol");
	desc = GetValueFromNameValueList(&data, "NewPortMappingDescription");
	leaseduration_str = GetValueFromNameValueList(&data,
						      "NewLeaseDuration");

	if ((int_port == NULL) || (ext_port == NULL) || (protocol == NULL)) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}
	if ((strcmp(protocol, "TCP") != 0) &&
	    (strcmp(protocol, "UDP") != 0)) {
	      ClearNameValueList(&data);
	      if (v2)
		      SoapError(h, 601, "Argument Value Out of Range");
	      else
		      SoapError(h, 402, "Invalid Args");
	      return;
	}

	eport = (unsigned short) atoi(ext_port);
	iport = (unsigned short) atoi(int_port);

	if (eport == 0) {
		ClearNameValueList(&data);
		SoapError(h, 716, "WildCardNotPermittedInExtPort");
		return;
	}

	if (iport == 0) {
		ClearNameValueList(&data);
		if (v2)
			SoapError(h, 732, "WildCardNotPermittedInIntPort");
		else
			SoapError(h, 402, "Invalid Args");
		return;
	}

	if (desc == NULL)
		desc = "b4iwf";

	if (leaseduration_str != NULL)
		leaseduration = (unsigned int) atoi(leaseduration_str);
	else
		leaseduration = 0U;
	/* PortMappingLeaseDuration can be either a value between 1 and
	 * 604800 seconds or the zero value (for infinite lease time).
	 * Note that an infinite lease time can be only set by out-of-band
	 * mechanisms like WWW-administration, remote management or local
	 * management.
	 * If a control point uses the value 0 to indicate an infinite lease
	 * time mapping, it is REQUIRED that gateway uses the maximum value
	 * instead (e.g. 604800 seconds) */
	if (v2 && ((leaseduration == 0U) || (leaseduration > 604800U)))
		leaseduration = 604800U;

	syslog(LOG_INFO,
	       "%s: ext port %hu to %s:%hu protocol %s for: "
	       "%s leaseduration=%u rhost=%s",
	       action, eport, int_ip, iport, protocol,
	       desc, leaseduration, r_host != NULL ? r_host : "*");

	upnp_dynamic_redirect(h, &eport, int_ip, iport, protocol,
			      desc, leaseduration, r_host, v2);

	ClearNameValueList(&data);
}

/* AddAnyPortMapping was added in WANIPConnection v2 */
static void
AddAnyPortMapping(struct upnphttp *h, const char *action, int v2)
{
	struct NameValueParserData data;
	const char *int_ip, *int_port, *ext_port, *protocol, *desc;
	const char *r_host, *leaseduration_str;
	unsigned short iport, eport;
	unsigned int leaseduration = 0U;
	struct hostent *hp;
	char **ptr;
	struct in_addr result_ip, remote;

	action = action;
	v2 = v2;
	if (external_addr == 0) {
		SoapError(h, 501, "ActionFailed");
		return;
	}

	ParseNameValue(h->req_buf + h->req_contentoff,
		       h->req_contentlen, &data);
	r_host = GetValueFromNameValueList(&data, "NewRemoteHost");
	ext_port = GetValueFromNameValueList(&data, "NewExternalPort");
	protocol = GetValueFromNameValueList(&data, "NewProtocol");
	int_port = GetValueFromNameValueList(&data, "NewInternalPort");
	int_ip = GetValueFromNameValueList(&data, "NewInternalClient");
	/* NewEnabled */
	desc = GetValueFromNameValueList(&data, "NewPortMappingDescription");
	leaseduration_str = GetValueFromNameValueList(&data,
						      "NewLeaseDuration");

	if ((int_port == NULL) || (ext_port == NULL) ||
	    (protocol == NULL) || (leaseduration_str == NULL)) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}
	if ((strcmp(protocol, "TCP") != 0) &&
	    (strcmp(protocol, "UDP") != 0)) {
	      ClearNameValueList(&data);
	      if (v2)
		      SoapError(h, 601, "Argument Value Out of Range");
	      else
		      SoapError(h, 402, "Invalid Args");
	      return;
	}

	if (leaseduration_str != NULL)
		leaseduration = (unsigned int) atoi(leaseduration_str);
	if ((leaseduration == 0U) || (leaseduration > 604800U))
		leaseduration = 604800U;

	eport = (unsigned short) atoi(ext_port);
	iport = (unsigned short) atoi(int_port);

	if (iport == 0) {
		ClearNameValueList(&data);
		SoapError(h, 732, "WildCardNotPermittedInIntPort");
		return;
	}

	if (int_ip == NULL) {
		ClearNameValueList(&data);
		SoapError(h, 715, "WildCardNotPermittedInSrcIP");
		return;
	}

	/* if ip not valid assume hostname and convert */
	if (inet_pton(AF_INET, int_ip, &result_ip) <= 0) {
		hp = gethostbyname(int_ip);
		if(hp && hp->h_addrtype == AF_INET) { 
			for(ptr = hp->h_addr_list;
			    (ptr != NULL) && (*ptr != NULL);
			    ptr++) {
				int_ip = inet_ntoa(*((struct in_addr *) *ptr));
				result_ip = *((struct in_addr *) *ptr);
				/* TODO : deal with more than
				 * one ip per hostname */
				break;
			}
		} 
		else {
			syslog(LOG_ERR,
			       "Failed to convert hostname '%s' to ip address",
			       int_ip); 
			ClearNameValueList(&data);
			SoapError(h, 402, "Invalid Args");
			return;
		}				
	}

	/* if rhost not valid assume hostname and convert */
	if (r_host != NULL) {
		if (inet_pton(AF_INET, r_host, &remote) <= 0) {
			hp = gethostbyname(r_host);
			if (hp && (hp->h_addrtype == AF_INET))
				for (ptr = hp->h_addr_list;
				     (ptr != NULL) && (*ptr != NULL);
				     ptr++) {
					remote = *((struct in_addr *) *ptr);
					r_host = inet_ntoa(remote);
					/* TODO : deal with more than
					   one ip per hostname */
					break;
				}
			else {
				syslog(LOG_ERR,
				       "Failed to convert hostname '%s'"
				       "to ip address",
				       r_host); 
				ClearNameValueList(&data);
				SoapError(h, 402, "Invalid Args");
				return;
			}
		}
	}

	/* check if NewInternalAddress is the client address */
	if(GETFLAG(SECUREMODEMASK))
		if(h->clientaddr.s_addr != result_ip.s_addr) {
			syslog(LOG_INFO,
			       "Client %s tried to redirect port to %s",
			       inet_ntoa(h->clientaddr), int_ip);
			ClearNameValueList(&data);
			SoapError(h, 606, "Action not authorized");
			return;
		}

	upnp_dynamic_redirect(h, &eport, int_ip, iport, protocol,
			      desc, leaseduration, r_host, -1);

	ClearNameValueList(&data);
}

static unsigned int
get_leaseduration(struct upnp_redirect *r)
{
	int ret;

	if (r->lifetime == 0U)
		return 0U;
	ret = r->expire.when.tv_sec - now.tv_sec;
	if (ret <= 0)
		return 1U;
	return (unsigned int) ret;
}

static void
GetSpecificPortMappingEntry(struct upnphttp *h, const char *action, int v2)
{
	struct upnp_redirect *r;
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s%d\">"
		"<NewInternalPort>%u</NewInternalPort>"
		"<NewInternalClient>%s</NewInternalClient>"
		"<NewEnabled>1</NewEnabled>"
		"<NewPortMappingDescription>%s</NewPortMappingDescription>"
		"<NewLeaseDuration>%u</NewLeaseDuration>"
		"</u:%sResponse>";
	char body[1024];
	int bodylen;
	struct NameValueParserData data;
	const char *r_host, *ext_port, *protocol;
	unsigned short eport;
	struct hostent *hp;
	char **ptr;
	struct in_addr remote;

	if (external_addr == 0) {
		SoapError(h, 501, "ActionFailed");
		return;
	}

	ParseNameValue(h->req_buf + h->req_contentoff,
		       h->req_contentlen, &data);
	r_host = GetValueFromNameValueList(&data, "NewRemoteHost");
	ext_port = GetValueFromNameValueList(&data, "NewExternalPort");
	protocol = GetValueFromNameValueList(&data, "NewProtocol");

	if ((ext_port == NULL) || (protocol == NULL)) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}
	if ((strcmp(protocol, "TCP") != 0) && (strcmp(protocol, "UDP") != 0)) {
		ClearNameValueList(&data);
		SoapError(h, 714, "NoSuchEntryInArray");
		return;
	}

	eport = (unsigned short) atoi(ext_port);

	/* if rhost not valid assume hostname and convert */
	if (r_host != NULL) {
		if (inet_pton(AF_INET, r_host, &remote) <= 0) {
			hp = gethostbyname(r_host);
			if (hp && (hp->h_addrtype == AF_INET))
				for (ptr = hp->h_addr_list;
				     (ptr != NULL) && (*ptr != NULL);
				     ptr++) {
					remote = *((struct in_addr *) *ptr);
					r_host = inet_ntoa(remote);
					/* TODO : deal with more than
					   one ip per hostname */
					break;
				}
			else {
				syslog(LOG_ERR,
				       "Failed to convert hostname '%s'"
				       "to ip address",
				       r_host); 
				ClearNameValueList(&data);
				SoapError(h, 402, "Invalid Args");
				return;
			}
		}
	}

	r = upnp_get_redirect(eport, protocol);

	if ((r == NULL) ||
	    (((r_host != NULL) && (strcmp(r_host, r->rhost) != 0)) ||
	     ((r_host == NULL) && (r->rhost[0] != 0))))
		SoapError(h, 714, "NoSuchEntryInArray");
	else {
		syslog(LOG_INFO,
		       "%s: rhost='%s' %s %s found => %s:%u desc='%s'",
		       action, r_host == NULL ? "*" : r_host,
		       ext_port, protocol,
		       r->iaddr, r->iport, r->desc);
		bodylen = snprintf(body, sizeof(body), resp, action,
				   SERVICE_TYPE_WANIPC,
				   v2 ? 2 : 1,
				   r->iport, r->iaddr, r->desc,
				   get_leaseduration(r), action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}

	ClearNameValueList(&data);
}

void
DelPortMappingHandler(struct upnphttp *h, int v2)
{
	static const char resp[] =
		"<u:DeletePortMappingResponse "
		"xmlns:u=\"" SERVICE_TYPE_WANIPC "%d\">"
		"</u:DeletePortMappingResponse>";
	char body[512];
	int bodylen;

	bodylen = snprintf(body, sizeof(body), resp, v2 ? 2 : 1);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static void
DeletePortMapping(struct upnphttp *h, const char *action, int v2)
{
	struct upnp_redirect *r;
	struct NameValueParserData data;
	const char *r_host, *ext_port, *protocol;
	unsigned short eport;
	struct in_addr iaddr;
	struct hostent *hp;
	char **ptr;
	struct in_addr remote;

	if (external_addr == 0) {
		SoapError(h, 501, "ActionFailed");
		return;
	}

	ParseNameValue(h->req_buf + h->req_contentoff,
		       h->req_contentlen, &data);
	r_host = GetValueFromNameValueList(&data, "NewRemoteHost");
	ext_port = GetValueFromNameValueList(&data, "NewExternalPort");
	protocol = GetValueFromNameValueList(&data, "NewProtocol");

	if ((ext_port == NULL) || (protocol == NULL)) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	} else if ((strcmp(protocol, "TCP") != 0) &&
		   (strcmp(protocol, "UDP") != 0)) {
	      ClearNameValueList(&data);
	      if (v2)
		      SoapError(h, 601, "Argument Value Out of Range");
	      else
		      SoapError(h, 714, "NoSuchEntryInArray");
	      return;
	}

	eport = (unsigned short) atoi(ext_port);

	/* if rhost not valid assume hostname and convert */
	if (r_host != NULL) {
		if (inet_pton(AF_INET, r_host, &remote) <= 0) {
			hp = gethostbyname(r_host);
			if (hp && (hp->h_addrtype == AF_INET))
				for (ptr = hp->h_addr_list;
				     (ptr != NULL) && (*ptr != NULL);
				     ptr++) {
					remote = *((struct in_addr *) *ptr);
					r_host = inet_ntoa(remote);
					/* TODO : deal with more than
					   one ip per hostname */
					break;
				}
			else {
				syslog(LOG_ERR,
				       "Failed to convert hostname '%s'"
				       "to ip address",
				       r_host); 
				ClearNameValueList(&data);
				SoapError(h, 402, "Invalid Args");
				return;
			}
		}
	}

	syslog(LOG_INFO,
	       "%s: external port: %hu, protocol: %s", 
	       action, eport, protocol);

	r = upnp_get_redirect(eport, protocol);

	if ((r == NULL) ||
	    (((r_host != NULL) && (strcmp(r_host, r->rhost) != 0)) ||
	     ((r_host == NULL) && (r->rhost[0] != 0)))) {
		ClearNameValueList(&data);
		SoapError(h, 714, "NoSuchEntryInArray");
		return;
	}

	if (GETFLAG(SECUREMODEMASK)) {
		if (inet_pton(AF_INET, r->iaddr, &iaddr) <= 0) {
			syslog(LOG_ERR, "%s: inet_pton", action);
			ClearNameValueList(&data);
			SoapError(h, 501, "ActionFailed");
			return;
		}
		if (h->clientaddr.s_addr != iaddr.s_addr) {
			syslog(LOG_INFO,
			       "Client %s tried to delete port to %s",
			       inet_ntoa(h->clientaddr), r->iaddr);
			ClearNameValueList(&data);
			if (v2)
				SoapError(h, 606, "Action not authorized");
			else
				SoapError(h, 501, "ActionFailed");
			return;
		}
	}

	if (upnp_delete_redirect(h, eport, protocol, v2) < 0)
		SoapError(h, 501, "ActionFailed");

	ClearNameValueList(&data);
}

/* DeletePortMappingRange was added in IGD spec v2 */
static void
DeletePortMappingRange(struct upnphttp * h, const char * action, int v2)
{
	struct upnp_redirect *r;
	static const char resp[] =
		"<u:DeletePortMappingRangeResponse "
		"xmlns:u=\"" SERVICE_TYPE_WANIPC "2\">"
		"</u:DeletePortMappingRangeResponse>";
	struct NameValueParserData data;
	const char * protocol, * start_port, * end_port, * manage;
	unsigned short startport, endport;
	unsigned int i, found;
	struct in_addr iaddr;

	/* possible errors :
	   606 - Action not authorized
	   730 - PortMappingNotFound
	   733 - InconsistentParameters
	 */
	v2 = v2;
	if (external_addr == 0) {
		SoapError(h, 501, "ActionFailed");
		return;
	}

	ParseNameValue(h->req_buf + h->req_contentoff,
		       h->req_contentlen, &data);
	start_port = GetValueFromNameValueList(&data, "NewStartPort");
	end_port = GetValueFromNameValueList(&data, "NewEndPort");
	protocol = GetValueFromNameValueList(&data, "NewProtocol");
	manage = GetValueFromNameValueList(&data, "NewManage");

	if ((start_port == NULL) || (end_port == NULL) ||
	    (protocol == NULL) || (manage == NULL)) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	} else if ((strcmp(protocol, "TCP") != 0) &&
		   (strcmp(protocol, "UDP") != 0)) {
	      ClearNameValueList(&data);
	      SoapError(h, 601, "Argument Value Out of Range");
	      return;
	}

	startport = (unsigned short)atoi(start_port);
	endport = (unsigned short)atoi(end_port);
	if (endport < startport) {
		ClearNameValueList(&data);
		SoapError(h, 733, "InconsistentParameters");
	}

	syslog(LOG_INFO,
	       "%s: start port: %hu, end port: %hu, protocol: %s, manage: %d",
	       action, startport, endport, protocol, atoi(manage));

	found = 0;
	for (i = startport; i <= endport; i++) {
		r = upnp_get_redirect((unsigned short) i, protocol);

		if (r == NULL)
			continue;
		if (inet_pton(AF_INET, r->iaddr, &iaddr) <= 0) {
			syslog(LOG_ERR, "%s: inet_pton", action);
			continue;
		}
		if (h->clientaddr.s_addr != iaddr.s_addr)
			if (GETFLAG(SECUREMODEMASK) ||
			    (atoi(manage) == 0))
				continue;
		if (upnp_delete_redirect(h, (unsigned short) i,
					 protocol, -1) < 0)
			continue;
		found++;
	}
	if (found == 0) {
		ClearNameValueList(&data);
		SoapError(h, 730, "PortMappingNotFound");
		return;
	}

	BuildSendAndCloseSoapResp(h, resp, sizeof(resp) - 1);
	ClearNameValueList(&data);
}

static void
GetGenericPortMappingEntry(struct upnphttp *h, const char *action, int v2)
{
	struct upnp_redirect *r;
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s%d\">"
		"<NewRemoteHost>%s</NewRemoteHost>"
		"<NewExternalPort>%hu</NewExternalPort>"
		"<NewProtocol>%s</NewProtocol>"
		"<NewInternalPort>%u</NewInternalPort>"
		"<NewInternalClient>%s</NewInternalClient>"
		"<NewEnabled>1</NewEnabled>"
		"<NewPortMappingDescription>%s</NewPortMappingDescription>"
		"<NewLeaseDuration>%u</NewLeaseDuration>"
		"</u:%sResponse>";
	int index = 0;
	const char *m_index;
	struct NameValueParserData data;

	if (external_addr == 0) {
		SoapError(h, 501, "ActionFailed");
		return;
	}

	ParseNameValue(h->req_buf + h->req_contentoff,
		       h->req_contentlen, &data);
	m_index = GetValueFromNameValueList(&data, "NewPortMappingIndex");

	if (m_index == NULL) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}	

	index = (int) atoi(m_index);

	syslog(LOG_INFO, "%s: index=%d", action, index);

	r = upnp_get_redirect_by_index(index);

	if (r == NULL)
		SoapError(h, 713, "SpecifiedArrayIndexInvalid");
	else {
		int bodylen;
		char body[2048];

		bodylen = snprintf(body, sizeof(body), resp, action,
				   SERVICE_TYPE_WANIPC, v2 ? 2 : 1,
				   r->rhost != NULL ? r->rhost : "",
				   r->eport,
				   r->proto == IPPROTO_UDP ? "UDP" : "TCP",
				   r->iport, r->iaddr, r->desc,
				   get_leaseduration(r), action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}

	ClearNameValueList(&data);
}

/* strcat_str()
 * concatenate the string and use realloc to increase the
 * memory buffer if needed. */
/* from upnpdescgen.c */
static char *
strcat_str(char * str, int * len, int * tmplen, const char * s2)
{
	int s2len;
	char * ostr = str;
	s2len = (int)strlen(s2);
	if(*tmplen <= (*len + s2len))
	{
		if(s2len < 256)
			*tmplen += 256;
		else
			*tmplen += s2len + 1;
		str = (char *)realloc(str, *tmplen);
		if (str == NULL)
		{
			free(ostr);
			return NULL;
		}
	}
	/*strcpy(str + *len, s2); */
	memcpy(str + *len, s2, s2len + 1);
	*len += s2len;
	return str;
}

/*
build the PortMappingList xml document :

<p:PortMappingList xmlns:p="urn:schemas-upnp-org:gw:WANIPConnection"
xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"
xsi:schemaLocation="urn:schemas-upnp-org:gw:WANIPConnection
http://www.upnp.org/schemas/gw/WANIPConnection-v2.xsd">
<p:PortMappingEntry>
<p:NewRemoteHost>202.233.2.1</p:NewRemoteHost>
<p:NewExternalPort>2345</p:NewExternalPort>
<p:NewProtocol>TCP</p:NewProtocol>
<p:NewInternalPort>2345</p:NewInternalPort>
<p:NewInternalClient>192.168.1.137</p:NewInternalClient>
<p:NewEnabled>1</p:NewEnabled>
<p:NewDescription>dooom</p:NewDescription>
<p:NewLeaseTime>345</p:NewLeaseTime>
</p:PortMappingEntry>
</p:PortMappingList>
*/

static char *
genEntry(struct upnp_redirect *r, char * str, int * len, int * tmplen)
{
	char tmp[2048];

	(void) snprintf(tmp, sizeof(tmp),
			"<p:PortMappingEntry>"
			"<p:NewRemoteHost>%s</p:NewRemoteHost>"
			"<p:NewExternalPort>%hu</p:NewExternalPort>"
			"<p:NewProtocol>%s</p:NewProtocol>"
			"<p:NewInternalPort>%u</p:NewInternalPort>"
			"<p:NewInternalClient>%s</p:NewInternalClient>"
			"<p:NewEnabled>1</p:NewEnabled>"
			"<p:NewDescription>%s</p:NewDescription>"
			"<p:NewLeaseTime>%u</p:NewLeaseTime>"
			"</p:PortMappingEntry>",
			r->rhost != NULL ? r->rhost : "",
			r->eport,
			r->proto == IPPROTO_UDP ? "UDP" : "TCP",
			r->iport, r->iaddr, r->desc,
			get_leaseduration(r));
	return strcat_str(str, len, tmplen, tmp);
}

/* GetListOfPortMappings was added in the IGD v2 specification */
static void
GetListOfPortMappings(struct upnphttp * h, const char * action, int v2)
{
	struct upnp_redirect *r;
	char * str;
	int len, tmplen;
	struct NameValueParserData data;
	const char * protocol, * start_port, * end_port, * manage, * number;
	unsigned short startport, endport;
	unsigned int limit, i, found;
	struct in_addr iaddr;

	v2 = v2;
	if (external_addr == 0) {
		SoapError(h, 501, "ActionFailed");
		return;
	}

	ParseNameValue(h->req_buf + h->req_contentoff,
		       h->req_contentlen, &data);
	start_port = GetValueFromNameValueList(&data, "NewStartPort");
	end_port = GetValueFromNameValueList(&data, "NewEndPort");
	protocol = GetValueFromNameValueList(&data, "NewProtocol");
	manage = GetValueFromNameValueList(&data, "NewManage");
	number = GetValueFromNameValueList(&data, "NewNumberOfPorts");

	if ((start_port == NULL) || (end_port == NULL) ||
	    (protocol == NULL) || (manage == NULL) || (number == NULL)) {
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	} else if ((strcmp(protocol, "TCP") != 0) &&
		   (strcmp(protocol, "UDP") != 0)) {
		ClearNameValueList(&data);
		SoapError(h, 601, "Argument Value Out of Range");
		return;
	}

	startport = (unsigned short)atoi(start_port);
	endport = (unsigned short)atoi(end_port);
	if (endport < startport) {
		ClearNameValueList(&data);
		SoapError(h, 733, "InconsistentParameters");
		return;
	}

	syslog(LOG_INFO,
	       "%s: start port: %hu, end port: %hu, protocol: %s, manage: %d",
	       action, startport, endport, protocol, atoi(manage));

	len = 0;
	tmplen = 2048;
	str = (char *) malloc(tmplen);
	if (str == NULL) {
		ClearNameValueList(&data);
		SoapError(h, 501, "ActionFailed");
		return;
	}
	str = strcat_str(str, &len, &tmplen,
		"<u:GetListOfPortMappingsResponse "
		"xmlns:u=\"" SERVICE_TYPE_WANIPC "2\">"
		"<NewPortListing><![CDATA["
		"<p:PortMappingList "
		"xmlns:p=\"urn:schemas-upnp-org:gw:WANIPConnection\" "
                "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" "
                "xsi:schemaLocation=\"urn:schemas-upnp-org:gw:WANIPConnection "
		"http://www.upnp.org/schemas/gw/WANIPConnection-v2.xsd\">");
	if (str == NULL) {
		ClearNameValueList(&data);
		SoapError(h, 501, "ActionFailed");
		return;
	}
	limit = (unsigned int)atoi(number);
#if 0
	if (limit == 0)
		limit = 1000;
#endif
	found = 0;
	for (i = startport; i <= endport; i++) {
		r = upnp_get_redirect((unsigned short) i, protocol);

		if (r == NULL)
			continue;
		if (inet_pton(AF_INET, r->iaddr, &iaddr) <= 0) {
			syslog(LOG_ERR, "%s: inet_pton", action);
			continue;
		}
		if (h->clientaddr.s_addr != iaddr.s_addr)
			if (GETFLAG(SECUREMODEMASK) ||
			    (atoi(manage) == 0))
				continue;
		str = genEntry(r, str, &len, &tmplen);
		if (str == NULL) {
			ClearNameValueList(&data);
			SoapError(h, 501, "ActionFailed");
			return;
		}
		found++;
		if (found == limit)
			break;
	}
	if (found == 0) {
		ClearNameValueList(&data);
		SoapError(h, 730, "PortMappingNotFound");
		return;
	}

	str = strcat_str(str, &len, &tmplen,
			 "</p:PortMappingList>"
			 "]]></NewPortListing>"
			 "</u:GetListOfPortMappingsResponse>");
	if (str == NULL) {
		ClearNameValueList(&data);
		SoapError(h, 501, "ActionFailed");
		return;
	}
	ClearNameValueList(&data);
	BuildSendAndCloseSoapResp(h, str, len);
}

#ifdef ENABLE_L3F_SERVICE
static void
SetDefaultConnectionService(struct upnphttp * h, const char * action, int v2)
{
	static const char resp[] =
		"<u:SetDefaultConnectionServiceResponse "
		"xmlns:u=\"urn:schemas-upnp-org:service:Layer3Forwarding:1\">"
		"</u:SetDefaultConnectionServiceResponse>";
	struct NameValueParserData data;
	char * p;

	v2 = v2;
	ParseNameValue(h->req_buf + h->req_contentoff,
		       h->req_contentlen, &data);
	p = GetValueFromNameValueList(&data, "NewDefaultConnectionService");
	if(p) {
		syslog(LOG_INFO, "%s(%s) : Ignored", action, p);
	}
	ClearNameValueList(&data);
	BuildSendAndCloseSoapResp(h, resp, sizeof(resp)-1);
}

static void
GetDefaultConnectionService(struct upnphttp * h, const char * action, int v2)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"urn:schemas-upnp-org:service:Layer3Forwarding:1\">"
		"<NewDefaultConnectionService>%s:WANConnectionDevice:1,"
		SERVICE_ID_WANIPC "</NewDefaultConnectionService>"
		"</u:%sResponse>";
	/* example from UPnP_IGD_Layer3Forwarding 1.0.pdf :
	 * uuid:44f5824f-c57d-418c-a131-f22b34e14111:WANConnectionDevice:1,
	 * urn:upnp-org:serviceId:WANPPPConn1 */
	char body[1024];
	int bodylen;

	v2 = v2;
	bodylen = snprintf(body, sizeof(body), resp,
	                   action, uuidvalue, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}
#endif

/* Added for compliance with WANIPConnection v2 */
static void
SetConnectionType(struct upnphttp * h, const char * action, int v2)
{
	const char * connection_type;
	struct NameValueParserData data;

	action = action;
	ParseNameValue(h->req_buf + h->req_contentoff,
		       h->req_contentlen, &data);
	connection_type = GetValueFromNameValueList(&data, "NewConnectionType");
	/* Unconfigured, IP_Routed, IP_Bridged */
	connection_type = connection_type;
	ClearNameValueList(&data);
	if (v2)
		SoapError(h, 731, "ReadOnly");
	else
		SoapError(h, 501, "ActionFailed");
}

/* Added for compliance with WANIPConnection v2 */
static void
RequestConnection(struct upnphttp * h, const char * action, int v2)
{
	action = action;
	if (v2)
		SoapError(h, 606, "Action not authorized");
	else
		SoapError(h, 501, "ActionFailed");
}

/* Added for compliance with WANIPConnection v2 */
static void
ForceTermination(struct upnphttp * h, const char * action, int v2)
{
	action = action;
	if (v2)
		SoapError(h, 606, "Action not authorized");
	else
		SoapError(h, 501, "ActionFailed");
}

/*
If a control point calls QueryStateVariable on a state variable that is not
buffered in memory within (or otherwise available from) the service,
the service must return a SOAP fault with an errorCode of 404 Invalid Var.

QueryStateVariable remains useful as a limited test tool but may not be
part of some future versions of UPnP.
*/
static void
QueryStateVariable(struct upnphttp * h, const char * action, int v2)
{
	static const char resp[] =
        "<u:%sResponse "
        "xmlns:u=\"%s\">"
		"<return>%s</return>"
        "</u:%sResponse>";

	char body[512];
	int bodylen;
	struct NameValueParserData data;
	const char * var_name;

	v2 = v2;
	ParseNameValue(h->req_buf + h->req_contentoff,
		       h->req_contentlen, &data);
	/*var_name = GetValueFromNameValueList(&data, "QueryStateVariable"); */
	/*var_name = GetValueFromNameValueListIgnoreNS(&data, "varName");*/
	var_name = GetValueFromNameValueList(&data, "varName");

	/*syslog(LOG_INFO, "QueryStateVariable(%.40s)", var_name); */

	if(!var_name)
	{
		SoapError(h, 402, "Invalid Args");
	}
	else if(strcmp(var_name, "ConnectionStatus") == 0)
	{	
		const char * status;

		status = get_wan_connection_status_str(ext_if_name);
		bodylen = snprintf(body, sizeof(body), resp,
                           action, "urn:schemas-upnp-org:control-1-0",
		                   status, action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}
	else if(strcmp(var_name, "PortMappingNumberOfEntries") == 0)
	{
		char strn[10];
		snprintf(strn, sizeof(strn), "%i",
		         upnp_get_redirect_number());
		bodylen = snprintf(body, sizeof(body), resp,
                           action, "urn:schemas-upnp-org:control-1-0",
		                   strn, action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}
	else if(strcmp(var_name, "SystemUpdateID") == 0)
	{
		char strn[10];
		snprintf(strn, sizeof(strn), "%lu", SystemUpdateID);
		bodylen = snprintf(body, sizeof(body), resp,
                           action, "urn:schemas-upnp-org:control-1-0",
		                   strn, action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}
#ifdef ENABLE_6FC_SERVICE
	else if(strcmp(var_name, "FirewallEnabled") == 0)
	{
		char strn[10];
		snprintf(strn, sizeof(strn), "%d", ipv6fc_firewall_enabled);
		bodylen = snprintf(body, sizeof(body), resp,
                           action, "urn:schemas-upnp-org:control-1-0",
		                   strn, action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}
	else if(strcmp(var_name, "InboundPinholeAllowed") == 0)
	{
		char strn[10];
		snprintf(strn, sizeof(strn), "%d",
			 ipv6fc_inbound_pinhole_allowed);
		bodylen = snprintf(body, sizeof(body), resp,
                           action, "urn:schemas-upnp-org:control-1-0",
		                   strn, action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}
#endif
	else
	{
		syslog(LOG_NOTICE, "%s: Unknown: %s", action, var_name?var_name:"");
		SoapError(h, 404, "Invalid Var");
	}

	ClearNameValueList(&data);	
}

#ifdef ENABLE_6FC_SERVICE
/* WANIPv6FirewallControl actions */
static void
GetFirewallStatus(struct upnphttp * h, const char * action)
{
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<FirewallEnabled>%d</FirewallEnabled>"
		"<InboundPinholeAllowed>%d</InboundPinholeAllowed>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;

	bodylen = snprintf(body, sizeof(body), resp,
		action, SERVICE_TYPE_WAN6FC,
		ipv6fc_firewall_enabled, ipv6fc_inbound_pinhole_allowed, action);
	BuildSendAndCloseSoapResp(h, body, bodylen);
}

static int
CheckStatus(struct upnphttp * h)
{
	if (!ipv6fc_firewall_enabled)
	{
		SoapError(h, 702, "FirewallDisabed");
		return 0;
	}
	else if(!ipv6fc_inbound_pinhole_allowed)
	{
		SoapError(h, 703, "InboundPinholeNotAllowed");
		return 0;
	}
	else
		return 1;
}

static int
DataVerification(struct upnphttp * h, char * int_ip, unsigned short * int_port, const char * protocol, char * leaseTime)
{
	//int n;
	// **  Internal IP can't be wildcarded
	if (!int_ip)
	{
		SoapError(h, 708, "WildCardNotPermittedInSrcIP");
		return 0;
	}
	
	if (!strchr(int_ip, ':'))
	{
		SoapError(h, 402, "Invalid Args");
		return 0;
	}

	// ** Internal port can't be wilcarded. 
//	printf("\tint_port: *%d*\n", *int_port);
	if (*int_port == 0)
	{
		SoapError(h, 706, "InternalPortWilcardingNotAllowed");
		return 0;
	}

	// ** Protocol can't be wilcarded and can't be an unknown port (here deal with only UDP, TCP, UDPLITE)
//	printf("\tprotocol: *%s*\n", protocol);
	if (atoi(protocol) == 65535)
	{
		SoapError(h, 707, "ProtocolWilcardingNotAllowed");
		return 0;
	}
	else if (atoi(protocol) != IPPROTO_UDP
	        && atoi(protocol) != IPPROTO_TCP
#ifdef IPPROTO_UDPITE
	        && atoi(protocol) != IPPROTO_UDPLITE
#endif
	        )
	{
		SoapError(h, 705, "ProtocolNotSupported");
		return 0;
	}

	// ** Lease Time can't be wilcarded nor >86400.
//	printf("\tlease time: %s\n", leaseTime);
	if(!leaseTime || !atoi(leaseTime) || atoi(leaseTime)>86400)
	{
		/* lease duration is never infinite, nor wilcarded. In this case, use default value */
		syslog(LOG_WARNING, "LeaseTime=%s not supported, (ip=%s)", leaseTime, int_ip);
		SoapError(h, 402, "Invalid Args");
		return 0;
	}

	return 1;
}

#if 0
static int connecthostport(const char * host, unsigned short port, char * result)
{
	int s, n;
	char hostname[INET6_ADDRSTRLEN];
	char port_str[8], ifname[8], tmp[4];
	struct addrinfo *ai, *p;
	struct addrinfo hints;

	memset(&hints, 0, sizeof(hints));
	/* hints.ai_flags = AI_ADDRCONFIG; */
#ifdef AI_NUMERICSERV
	hints.ai_flags = AI_NUMERICSERV;
#endif
	hints.ai_socktype = SOCK_STREAM;
	hints.ai_family = AF_UNSPEC; /* AF_INET, AF_INET6 or AF_UNSPEC */
	/* hints.ai_protocol = IPPROTO_TCP; */
	snprintf(port_str, sizeof(port_str), "%hu", port);
	strcpy(hostname, host);
	if(!strncmp(host, "fe80", 4))
	{
		printf("Using an linklocal address\n");
		strcpy(ifname, "%");
		snprintf(tmp, sizeof(tmp), "%d", linklocal_index);
		strcat(ifname, tmp);
		strcat(hostname, ifname);
		printf("host: %s\n", hostname);
	}
	n = getaddrinfo(hostname, port_str, &hints, &ai);
	if(n != 0)
	{
		fprintf(stderr, "getaddrinfo() error : %s\n", gai_strerror(n));
		return -1;
	}
	s = -1;
	for(p = ai; p; p = p->ai_next)
	{
#ifdef DEBUG
		char tmp_host[256];
		char tmp_service[256];
		printf("ai_family=%d ai_socktype=%d ai_protocol=%d ai_addrlen=%d\n ",
		       p->ai_family, p->ai_socktype, p->ai_protocol, p->ai_addrlen);
		getnameinfo(p->ai_addr, p->ai_addrlen, tmp_host, sizeof(tmp_host),
		            tmp_service, sizeof(tmp_service),
		            NI_NUMERICHOST | NI_NUMERICSERV);
		printf(" host=%s service=%s\n", tmp_host, tmp_service);
#endif
		inet_ntop(AF_INET6, &(((struct sockaddr_in6 *)p->ai_addr)->sin6_addr), result, INET6_ADDRSTRLEN);
		return 0;
	}
	freeaddrinfo(ai);
}
#endif

/* Check the security policy right */
static int
PinholeVerification(struct upnphttp * h, char * int_ip, unsigned short * int_port)
{
	int n;
	/* Pinhole InternalClient address must correspond to the action sender */
	syslog(LOG_INFO, "Checking internal IP@ and port (Security policy purpose)");
	char senderAddr[INET6_ADDRSTRLEN]="";
	//char str[INET6_ADDRSTRLEN]="";
	//connecthostport(int_ip, *int_port, str);
	//printf("int_ip: %s / str: %s\n", int_ip, str);
	
	struct addrinfo hints, *ai, *p;
	struct in6_addr result_ip;/*unsigned char result_ip[16];*/ /* inet_pton() */ //IPv6 Modification

	hints.ai_socktype = SOCK_STREAM;
	hints.ai_family = AF_UNSPEC;

	/* if ip not valid assume hostname and convert */
	if (inet_pton(AF_INET6, int_ip, &result_ip) <= 0) //IPv6 Modification
	{

		n = getaddrinfo(int_ip, NULL, &hints, &ai);//hp = gethostbyname(int_ip);
		if(!n && ai->ai_family == AF_INET6) //IPv6 Modification
		{
			for(p = ai; p; p = p->ai_next)//ptr = hp->h_addr_list; ptr && *ptr; ptr++)
		   	{
				inet_ntop(AF_INET6, (struct in6_addr *) p, int_ip, sizeof(struct in6_addr)); ///IPv6 Modification
				result_ip = *((struct in6_addr *) p);
				fprintf(stderr, "upnpsoap / AddPinhole: assuming int addr = %s", int_ip);
				/* TODO : deal with more than one ip per hostname */
				break;
			}
		}
		else
		{
			syslog(LOG_ERR, "Failed to convert hostname '%s' to ip address", int_ip);
			SoapError(h, 402, "Invalid Args");
			return -1;
		}
        freeaddrinfo(p);
	}

	if(inet_ntop(AF_INET6, &(h->clientaddr_v6), senderAddr, INET6_ADDRSTRLEN)<=0)
	{
		//printf("Failed to inet_ntop\n");
		syslog(LOG_ERR, "inet_ntop: %m");
	}
#ifdef DEBUG
	printf("\tPinholeVerification:\n\t\tCompare sender @: %s\n\t\t  to intClient @: %s\n", senderAddr, int_ip);
#endif
	if(strcmp(senderAddr, int_ip) != 0)
	if(h->clientaddr_v6.s6_addr != result_ip.s6_addr)
	{
		syslog(LOG_INFO, "Client %s tried to access pinhole for internal %s and is not authorized to do it",
		       senderAddr, int_ip);
		SoapError(h, 606, "Action not authorized");
		return 0;
	}

	/* Pinhole InternalPort must be greater than or equal to 1024 */
	if (*int_port < 1024)
	{
		syslog(LOG_INFO, "Client %s tried to access pinhole with port < 1024 and is not authorized to do it",
		       senderAddr);
		SoapError(h, 606, "Action not authorized");
		return 0;
	}
	return 1;
}

static void
AddPinhole(struct upnphttp * h, const char * action)
{
	int r;
	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<UniqueID>%d</UniqueID>"
		"</u:%sResponse>";
	char body[512];
	int bodylen;
	struct NameValueParserData data;
	char * rem_host, * rem_port, * int_ip, * int_port, * protocol, * leaseTime;
	int uid = 0;
	unsigned short iport, rport;

	if(CheckStatus(h)==0)
		return;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	rem_host = GetValueFromNameValueList(&data, "RemoteHost");
	rem_port = GetValueFromNameValueList(&data, "RemotePort");
	int_ip = GetValueFromNameValueList(&data, "InternalClient");
	int_port = GetValueFromNameValueList(&data, "InternalPort");
	protocol = GetValueFromNameValueList(&data, "Protocol");
	leaseTime = GetValueFromNameValueList(&data, "LeaseTime");

	rport = (unsigned short)atoi(rem_port);
	iport = (unsigned short)atoi(int_port);

	// **  As there is no security policy, InternalClient must be equal to the CP's IP address.
	if(DataVerification(h, int_ip, &iport, protocol, leaseTime) == 0
	   || PinholeVerification(h, int_ip, &iport) <= 0)
	{
		ClearNameValueList(&data);
		return ;
	}

	// ** RemoteHost can be wilcarded or an IDN.
	/*printf("\trem_host: %s\n", rem_host);*/
	if (rem_host!=NULL && !strchr(rem_host, ':'))
	{
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}
	/*printf("\tAddr check passed.\n");*/

	syslog(LOG_INFO, "%s: (inbound) from [%s]:%hu to [%s]:%hu with protocol %s during %ssec", action, rem_host?rem_host:"anywhere", rport, int_ip, iport, protocol, leaseTime);

	r = upnp_add_inboundpinhole(rem_host, rport, int_ip, iport, protocol, leaseTime, &uid);

	switch(r)
	{
		case 1:	        /* success */
			bodylen = snprintf(body, sizeof(body), resp, action, SERVICE_TYPE_WAN6FC, uid, action);
			BuildSendAndCloseSoapResp(h, body, bodylen);
			break;
		case -1: 	/* not permitted */
			SoapError(h, 701, "PinholeSpaceExhausted");
			break;
		default:
			SoapError(h, 501, "ActionFailed");
			break;
	}
	ClearNameValueList(&data);
}

static void
UpdatePinhole(struct upnphttp * h, const char * action)
{
	int r, n;
	static const char resp[] =
		"<u:UpdatePinholeResponse "
		"xmlns:u=\"" SERVICE_TYPE_WAN6FC "\">"
		"</u:UpdatePinholeResponse>";
	struct NameValueParserData data;
	const char * uid, * leaseTime;
	char iaddr[40], proto[6], lt[12];
	unsigned short iport;

	if(CheckStatus(h)==0)
		return;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	uid = GetValueFromNameValueList(&data, "UniqueID");
	leaseTime = GetValueFromNameValueList(&data, "NewLeaseTime");

	if(!uid || !leaseTime || !atoi(leaseTime) || atoi(leaseTime) > 86400)
	{
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	// Check that client is not deleting an pinhole he doesn't have access to, because of its public access
	n = upnp_get_pinhole_info(0, 0, iaddr, &iport, proto, uid, lt);
	if (n > 0)
	{
		if(PinholeVerification(h, iaddr, &iport)==0)
		{
			ClearNameValueList(&data);
			return ;
		}
	}

	syslog(LOG_INFO, "%s: (inbound) updating lease duration to %s for pinhole with ID: %s", action, leaseTime, uid);

	r = upnp_update_inboundpinhole(uid, leaseTime);

	if(r < 0)
	{
		if(r == -4 || r == -1)
			SoapError(h, 704, "NoSuchEntry");
		else
			SoapError(h, 501, "ActionFailed");
	}
	else
	{
		BuildSendAndCloseSoapResp(h, resp, sizeof(resp)-1);
	}
	ClearNameValueList(&data);
}

static void
GetOutboundPinholeTimeout(struct upnphttp * h, const char * action)
{
	if (!ipv6fc_firewall_enabled)
	{
		SoapError(h, 702, "FirewallDisabed");
		return;
	}
	int r;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<OutboundPinholeTimeout>%d</OutboundPinholeTimeout>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct NameValueParserData data;
	char * int_ip, * int_port, * rem_host, * rem_port, * protocol;
	int opt=0, proto=0;
	unsigned short iport, rport;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	int_ip = GetValueFromNameValueList(&data, "InternalClient");
	int_port = GetValueFromNameValueList(&data, "InternalPort");
	rem_host = GetValueFromNameValueList(&data, "RemoteHost");
	rem_port = GetValueFromNameValueList(&data, "RemotePort");
	protocol = GetValueFromNameValueList(&data, "Protocol");

	rport = (unsigned short)atoi(rem_port);
	iport = (unsigned short)atoi(int_port);
	proto = atoi(protocol);

	syslog(LOG_INFO, "%s: retrieving timeout for outbound pinhole from [%s]:%hu to [%s]:%hu protocol %s", action, int_ip, iport,rem_host, rport, protocol);

	r = upnp_check_outbound_pinhole(proto, &opt);

	switch(r)
	{
		case 1:	/* success */
			bodylen = snprintf(body, sizeof(body), resp, action, SERVICE_TYPE_WAN6FC, opt, action);
			BuildSendAndCloseSoapResp(h, body, bodylen);
			break;
		case -5:	/* Protocol not supported */
			SoapError(h, 705, "ProtocolNotSupported");
			break;
		default:
			SoapError(h, 501, "ActionFailed");
	}
	ClearNameValueList(&data);
}

static void
DeletePinhole(struct upnphttp * h, const char * action)
{
	if(CheckStatus(h)==0)
		return;
	int r, n;

	static const char resp[] =
		"<u:DeletePinholeResponse "
		"xmlns:u=\"" SERVICE_TYPE_WAN6FC "\">"
		"</u:DeletePinholeResponse>";

	struct NameValueParserData data;
	const char * uid;
	char iaddr[40], proto[6], lt[12];
	unsigned short iport;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	uid = GetValueFromNameValueList(&data, "UniqueID");

	if(!uid)
	{
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	// Check that client is not deleting an pinhole he doesn't have access to, because of its public access
	n = upnp_get_pinhole_info(0, 0, iaddr, &iport, proto, uid, lt);
	if (n > 0)
	{
		if(PinholeVerification(h, iaddr, &iport)==0)
		{
			ClearNameValueList(&data);
			return ;
		}
	}

	syslog(LOG_INFO, "%s: (inbound) delete pinhole with ID: %s", action, uid);

	r = upnp_delete_inboundpinhole(uid);

	if(r <= 0)
	{
		syslog(LOG_INFO, "%s: (inbound) failed to remove pinhole with ID: %s", action, uid);
		if(r==-4)
			SoapError(h, 704, "NoSuchEntry");
		else
			SoapError(h, 501, "ActionFailed");
	}
	else
	{
		syslog(LOG_INFO, "%s: (inbound) pinhole successfully removed", action);
		BuildSendAndCloseSoapResp(h, resp, sizeof(resp)-1);
	}
	ClearNameValueList(&data);
}

static void
CheckPinholeWorking(struct upnphttp * h, const char * action)
{
	if(CheckStatus(h)==0)
		return;
	int r, d;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<IsWorking>%d</IsWorking>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct NameValueParserData data;
	const char * uid;
	char eaddr[40], iaddr[40], proto[6], lt[12];
	unsigned short eport, iport;
	int isWorking = 0;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	uid = GetValueFromNameValueList(&data, "UniqueID");

	if(!uid)
	{
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	// Check that client is not checking a pinhole he doesn't have access to, because of its public access
	r = upnp_get_pinhole_info(eaddr, eport, iaddr, &iport, proto, uid, lt);
	if (r > 0)
	{
		if(PinholeVerification(h, iaddr, &iport)==0)
		{
			ClearNameValueList(&data);
			return ;
		}
		else
		{
			int rulenum_used, rulenum = 0;
			d = upnp_check_pinhole_working(uid, eaddr, iaddr, &eport, &iport, proto, &rulenum_used);
			if(d < 0)
			{
				if(d == -4)
				{
					syslog(LOG_INFO, "%s: rule for ID=%s, no trace found for this pinhole", action, uid);
					SoapError(h, 709, "NoPacketSent");
					ClearNameValueList(&data);
					return ;
				}
				else
				{
				// d==-5 not same table // d==-6 not same chain // d==-7 not found a rule but policy traced
					isWorking=0;
					syslog(LOG_INFO, "%s: rule for ID=%s is not working, packet going through %s", action, uid, (d==-5)?"the wrong table":((d==-6)?"the wrong chain":"a chain policy"));
					bodylen = snprintf(body, sizeof(body), resp,
						action, SERVICE_TYPE_WAN6FC,
						isWorking, action);
					BuildSendAndCloseSoapResp(h, body, bodylen);
				}
			}
			else
			{
				/*check_rule_from_file(uid, &rulenum);*/
				if(rulenum_used == rulenum)
				{
					isWorking=1;
					syslog(LOG_INFO, "%s: rule for ID=%s is working properly", action, uid);
				}
				else
				{
					isWorking=0;
					syslog(LOG_INFO, "%s: rule for ID=%s is not working", action, uid);
				}
				bodylen = snprintf(body, sizeof(body), resp,
						action, SERVICE_TYPE_WAN6FC,
						isWorking, action);
				BuildSendAndCloseSoapResp(h, body, bodylen);
			}
		}
	}
	else if(r == -4 || r == -1)
	{
		SoapError(h, 704, "NoSuchEntry");
	}
	else
	{
		SoapError(h, 501, "ActionFailed");
		ClearNameValueList(&data);
		return ;
	}
	ClearNameValueList(&data);
}

static void
GetPinholePackets(struct upnphttp * h, const char * action)
{
	if(CheckStatus(h)==0)
		return;
	int r, n;

	static const char resp[] =
		"<u:%sResponse "
		"xmlns:u=\"%s\">"
		"<PinholePackets>%d</PinholePackets>"
		"</u:%sResponse>";

	char body[512];
	int bodylen;
	struct NameValueParserData data;
	const char * uid;
	char iaddr[40], proto[6], lt[12];
	unsigned short iport;
	int pinholePackets = 0;

	ParseNameValue(h->req_buf + h->req_contentoff, h->req_contentlen, &data);
	uid = GetValueFromNameValueList(&data, "UniqueID");

	if(!uid)
	{
		ClearNameValueList(&data);
		SoapError(h, 402, "Invalid Args");
		return;
	}

	// Check that client is not getting infos of a pinhole he doesn't have access to, because of its public access
	r = upnp_get_pinhole_info(0, 0, iaddr, &iport, proto, uid, lt);
	if (r > 0)
	{
		if(PinholeVerification(h, iaddr, &iport)==0)
		{
			ClearNameValueList(&data);
			return ;
		}
	}

	n = upnp_get_pinhole_packets(uid, &pinholePackets);
	if(n > 0)
	{
		bodylen = snprintf(body, sizeof(body), resp,
				action, SERVICE_TYPE_WAN6FC,
				pinholePackets, action);
		BuildSendAndCloseSoapResp(h, body, bodylen);
	}
	else if(r == -4 || r == -1)
	{
		SoapError(h, 704, "NoSuchEntry");
	}
	else
	{
		SoapError(h, 501, "ActionFailed");
		ClearNameValueList(&data);
		return ;
	}
	ClearNameValueList(&data);
}
#endif


/* Windows XP as client send the following requests :
 * GetConnectionTypeInfo
 * GetNATRSIPStatus
 * ? GetTotalBytesSent - WANCommonInterfaceConfig
 * ? GetTotalBytesReceived - idem
 * ? GetTotalPacketsSent - idem
 * ? GetTotalPacketsReceived - idem
 * GetCommonLinkProperties - idem
 * GetStatusInfo - WANIPConnection
 * GetExternalIPAddress
 * QueryStateVariable / ConnectionStatus!
 */
static const struct 
{
	const char * methodName; 
	void (*methodImpl)(struct upnphttp *, const char *, int);
}
soapMethods[] =
{
	/* WANCommonInterfaceConfig */
	{ "QueryStateVariable", QueryStateVariable},
	{ "GetTotalBytesSent", GetTotalBytesSent},
	{ "GetTotalBytesReceived", GetTotalBytesReceived},
	{ "GetTotalPacketsSent", GetTotalPacketsSent},
	{ "GetTotalPacketsReceived", GetTotalPacketsReceived},
	{ "GetCommonLinkProperties", GetCommonLinkProperties},
	{ "GetStatusInfo", GetStatusInfo},
	/* WANIPConnection */
	{ "GetConnectionTypeInfo", GetConnectionTypeInfo },
	{ "GetNATRSIPStatus", GetNATRSIPStatus},
	{ "GetExternalIPAddress", GetExternalIPAddress},
	{ "AddPortMapping", AddPortMapping},
	{ "GetGenericPortMappingEntry", GetGenericPortMappingEntry},
	{ "GetSpecificPortMappingEntry", GetSpecificPortMappingEntry},
/* Required in WANIPConnection:2 */
	{ "SetConnectionType", SetConnectionType},
	{ "RequestConnection", RequestConnection},
	{ "ForceTermination", ForceTermination},
	{ "AddAnyPortMapping", AddAnyPortMapping},
	{ "DeletePortMappingRange", DeletePortMappingRange},
	{ "GetListOfPortMappings", GetListOfPortMappings},
/* Must be after DeletePortMappingRange */
	{ "DeletePortMapping", DeletePortMapping},
#ifdef ENABLE_L3F_SERVICE
	/* Layer3Forwarding */
	{ "SetDefaultConnectionService", SetDefaultConnectionService},
	{ "GetDefaultConnectionService", GetDefaultConnectionService},
#endif
#ifdef ENABLE_6FC_SERVICE
	/* WANIPv6FirewallControl */
	{ "GetFirewallStatus", GetFirewallStatus},
	{ "AddPinhole", AddPinhole},
	{ "UpdatePinhole", UpdatePinhole},
	{ "GetOutboundPinholeTimeout", GetOutboundPinholeTimeout},
	{ "DeletePinhole", DeletePinhole},
	{ "CheckPinholeWorking", CheckPinholeWorking},
	{ "GetPinholePackets", GetPinholePackets},
#endif
	{ 0, 0 }
};

void
ExecuteSoapAction(struct upnphttp * h, const char * action, int n)
{
	char * p;
	char * p2;
	int i, v2, len, methodlen;

	i = 0;
	v2 = 0;
	p = strchr(action, '#');

	if(p)
	{
		if ((p > action) && (p[-1] == '2'))
			v2 = 1;
		p++;
		p2 = strchr(p, '"');
		if(p2)
			methodlen = p2 - p;
		else
			methodlen = n - (p - action);
		/*syslog(LOG_DEBUG, "SoapMethod: %.*s", methodlen, p);*/
		while(soapMethods[i].methodName)
		{
			len = strlen(soapMethods[i].methodName);
			if(strncmp(p, soapMethods[i].methodName, len) == 0)
			{
				soapMethods[i].methodImpl(h, soapMethods[i].methodName, v2);
				return;
			}
			i++;
		}

		syslog(LOG_NOTICE, "SoapMethod: Unknown: %.*s", methodlen, p);
	}

	SoapError(h, 401, "Invalid Action");
}

/* Standard Errors:
 *
 * errorCode errorDescription Description
 * --------	---------------- -----------
 * 401 		Invalid Action 	No action by that name at this service.
 * 402 		Invalid Args 	Could be any of the following: not enough in args,
 * 							too many in args, no in arg by that name, 
 * 							one or more in args are of the wrong data type.
 * 403 		Out of Sync 	Out of synchronization.
 * 501 		Action Failed 	May be returned in current state of service
 * 							prevents invoking that action.
 * 600-699 	TBD 			Common action errors. Defined by UPnP Forum
 * 							Technical Committee.
 * 700-799 	TBD 			Action-specific errors for standard actions.
 * 							Defined by UPnP Forum working committee.
 * 800-899 	TBD 			Action-specific errors for non-standard actions. 
 * 							Defined by UPnP vendor.
*/
void
SoapError(struct upnphttp * h, int errCode, const char * errDesc)
{
	static const char resp[] = 
		"<s:Envelope "
		"xmlns:s=\"http://schemas.xmlsoap.org/soap/envelope/\" "
		"s:encodingStyle=\"http://schemas.xmlsoap.org/soap/encoding/\">"
		"<s:Body>"
		"<s:Fault>"
		"<faultcode>s:Client</faultcode>"
		"<faultstring>UPnPError</faultstring>"
		"<detail>"
		"<UPnPError xmlns=\"urn:schemas-upnp-org:control-1-0\">"
		"<errorCode>%d</errorCode>"
		"<errorDescription>%s</errorDescription>"
		"</UPnPError>"
		"</detail>"
		"</s:Fault>"
		"</s:Body>"
		"</s:Envelope>";

	char body[2048];
	int bodylen;

	syslog(LOG_INFO, "Returning UPnPError %d: %s", errCode, errDesc);
	bodylen = snprintf(body, sizeof(body), resp, errCode, errDesc);
	BuildResp2_upnphttp(h, 500, "Internal Server Error", body, bodylen);
	SendResp_upnphttp(h);
	CloseSocket_upnphttp(h);
}

