#!/usr/bin/perl -s

# $Id: 858f922b3b57933f6540fc58f6b768efdabe687d $

use POSIX;
use Socket;

open(TUN, "db_dump tunnels.db|") or
    open(TUN, "< tunnels.dump") or
    die "can't get tunnels";

my $key;
my $data;
my $mode = 0;

while (<TUN>) {
    next if $_ !~ /^ /;
    # $key = IPv6:16
    # $data = IPv4:4 + start:4
    if ($mode == 0) {
	$key = pack("H*", substr($_, 1));
	$mode = 1;
    } else {
	$data = pack("H*", substr($_, 1));
	$mode = 0;
	my $addr = substr($data, 0, 4);
	my $start = unpack("l", substr($data, 4));
	my $host6 = Socket::inet_ntop(AF_INET6, $key);
	my $host4 = Socket::inet_ntop(AF_INET, $addr);
	my $date = asctime(gmtime($start));
	printf("tunnel %s %s # %s", $host6, $host4, $date);
    }
}
undef $mode;
close(TUN);

open(MAP, "db_dump maps.db|") or
    open(MAP, "< maps.dump") or
    die "can't get static nats";

my $mode = 0;

while (<MAP>) {
    next if $_ !~ /^ /;
    # $key = IPv4:4 + nport:2 + proto:1
    # $data = IPv6:16 + IPv4/src:4 + IPv4/filter:4 + sport:2 +
    #    fport:2 + expire:4 + fplen:1 + fstate:1 + down:1
    if ($mode == 0) {
	$key = pack("H*", substr($_, 1));
	$mode = 1;
    } else {
	$data = pack("H*", substr($_, 1));
	$mode = 0;
	my $addr = substr($key, 0, 4);
	my ($nport, $proto) = unpack("S C", substr($key, 4));
	my $src = substr($data, 16, 4);
	my $faddr = substr($data, 20, 4);
	my ($sport, $fport, $expire) = unpack("S S l", substr($data, 24, 8));
	my ($fplen, $ftate, $down) = unpack("C c c", substr($data, 32));
	my $host6 = Socket::inet_ntop(AF_INET6, substr($data, 0, 16));
	my $nat4 = Socket::inet_ntop(AF_INET, $addr);
	my $src4 = Socket::inet_ntop(AF_INET, $src);
	my $faddr4 = Socket::inet_ntop(AF_INET, $faddr);
	my $p;
	if ($proto == Socket::IPPROTO_TCP) {
	    $p = "tcp";
	} elsif ($proto == Socket::IPPROTO_UDP) {
	    $p = "udp";
	} else {
	    $p = sprintf("proto%d", $proto);
	}
	my $s;
	if ($fstate == 0) {
	    $s = sprintf("map %s %s %s %hu %s %hu",
			 $host6, $p, $src4, $sport, $nat4, $nport);
	} else {
	    $s = sprintf("map %s %s %s %hu %s %hu [%s/%d %hu]",
			 $host6, $p, $src4, $sport, $nat4, $nport,
			 $faddr4, $fplen, $fport);
	}
	if ($expire == 0) {
	    if ($down != 0) {
		printf("%s # (down)", $s);
	    } else {
		printf("%s", $s);
	    }
	} else {
	    my $date = asctime(gmtime($expire));
	    if ($down != 0) {
		printf("%s # (down) %s", $s, $date);
	    } else {
		printf("%s # %s", $s, $date);
	    }
	}
    }
}
undef $mode;
close(MAP);
