#!/usr/bin/env python

# Copyright (C) 2012  Internet Systems Consortium, Inc. ("ISC")
#
# Permission to use, copy, modify, and/or distribute this software for any
# purpose with or without fee is hereby granted, provided that the above
# copyright notice and this permission notice appear in all copies.
#
# THE SOFTWARE IS PROVIDED "AS IS" AND ISC DISCLAIMS ALL WARRANTIES WITH
# REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY
# AND FITNESS.  IN NO EVENT SHALL ISC BE LIABLE FOR ANY SPECIAL, DIRECT,
# INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM
# LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE
# OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
# PERFORMANCE OF THIS SOFTWARE.
#

"""
Generate static mappings for a SD-AFTR

Francis_Dupont@isc.org, February 2012

Usage: IPv6_address IPv4_address Index Protocol BasePort NbPort

Constraints:
 Index 1..63
 Protocol tcp, udp or icmp
 BasePort 1..64512
 NbPort 1..1024

Algorithm:
    Index, BasePort + N <-> (N & 0xfff8) << 9 | Index << 3 | N & 0x0007
"""

__version__ = "$Id: 3c4aa6c1bbe7218906e4fee0990515f221591073 $"

import sys

def main(args):
    """main"""
    if (len(args) != 7):
        print 'Usage: IPv6_address IPv4_address Index Protocol BasePort NbPort'
        sys.exit(1)
    addr6 = args[1]
    addr4 = args[2]
    try:
        idx = int(args[3])
        if (idx < 1) or (idx > 63):
            raise ValueError
    except ValueError:
        print 'Index (', args[3], ') must be 1..63'
        sys.exit(1)
    proto = args[4]
    if (proto != 'tcp') and (proto != 'udp') and (proto != 'icmp'):
        print 'Protocol (', args[4], ') must be tcp, udp or icmp'
        sys.exit(1)
    try:
        base = int(args[5])
        if (base < 1) or (base > 64512):
            raise ValueError
    except ValueError:
        print 'BasePort (', args[5], ') must be 1..64512'
        sys.exit(1)
    try:
        nbp = int(args[6])
        if (nbp < 1) or (nbp > 1024):
            raise ValueError
    except ValueError:
        print 'NbPort (', args[6], ') must be 1..1024'
        sys.exit(1)
    print '# sdgen ', addr6, addr4, idx, proto, base, nbp
    print 'tunnel', addr6, addr4
    for i in xrange(0, nbp):
        port = i & 0x0007
        port |= idx << 3
        port |= (i & 0xfff8) << 9
        print 'nat', addr6, proto, addr4, base + i, addr4, port
    print '#'

main(sys.argv)
