/*
 * netfilter module to support ICMPv4 ID (from ip_tables.c, GPL)
 */
#include <linux/in.h>
#include <linux/icmp.h>
#include <linux/module.h>
#include <linux/skbuff.h>
#include <linux/netfilter/x_tables.h>
#include <net/netfilter/nf_conntrack.h>
#include <net/netfilter/nf_conntrack_core.h>
#include <net/netfilter/nf_conntrack_tuple.h>

#include "xt_echo.h"

static bool
echo_mt(const struct sk_buff *skb, struct xt_action_param *par)
{
	const struct icmphdr *ic;
	struct icmphdr _icmph;
	uint16_t _id;
	bool invert;
	const struct xt_echo_info *info = par->matchinfo;

	/* Must not be a fragment. */
	if (par->fragoff != 0)
		return false;

	ic = skb_header_pointer(skb, par->thoff, sizeof(_icmph), &_icmph);
	if (ic == NULL) {
		par->hotdrop = true;
		return false;
	}

	invert = !!(info->invflags & XT_ECHO_INV_ID);
	_id = ntohs(ic->un.echo.id);
	return (ic->type == info->type &&
		_id >= info->id[0] && _id <= info->id[1])
		^ invert;
}

static int
echo_mt_check(const struct xt_mtchk_param *par)
{
	struct xt_echo_info *info = par->matchinfo;

	if (info->type != ICMP_ECHO && info->type != ICMP_ECHOREPLY)
		return false;
	return !(info->invflags & ~XT_ECHO_INV_MASK);
}

static struct xt_match echo_mt_reg[] __read_mostly = {
	{
		.name       = "echo",
		.family     = AF_INET,
		.checkentry = echo_mt_check,
		.match      = echo_mt,
		.matchsize  = sizeof(struct xt_echo_info),
		.proto      = IPPROTO_ICMP,
		.me         = THIS_MODULE,
	},
};

static int __init echo_mt_init(void)
{
	return xt_register_matches(echo_mt_reg, ARRAY_SIZE(echo_mt_reg));
}

static void __exit echo_mt_exit(void)
{
	xt_unregister_matches(echo_mt_reg, ARRAY_SIZE(echo_mt_reg));
}

module_init(echo_mt_init);
module_exit(echo_mt_exit);
MODULE_AUTHOR("Francis Dupont");
MODULE_DESCRIPTION("Xtables: ICMPv4 echo ID");
MODULE_LICENSE("GPL");
MODULE_ALIAS("ipt_echo");
